"""
Pre-generate HTML and SVG element to `ptag/elements.py`

We take pre-generating approach because `Pylance` cannot infer the type of partial functions dynamically,
failing attempts included:

- update local variable with `locals().update()`
- use `exec()` to create partial functions.
"""

from pathlib import Path
from functools import partial
from typing import TypeVar, Callable, ParamSpec, Concatenate, cast

P = ParamSpec("P")
T = TypeVar("T")


def typed_partial(cls: Callable[Concatenate[str, P], T], tag_name: str) -> Callable[P, T]:
    return cast(Callable[P, T], partial(cls, tag_name))


html_tags = """\
a abbr acronym address applet area article aside audio b base basefont bdi bdo big blockquote body br button \
canvas caption center cite code col colgroup data datalist dd del details dfn dialog dir div dl dt em embed \
fieldset figcaption figure font footer form frame frameset h1 h2 h3 h4 h5 h6 head header hgroup hr html \
i iframe img input ins kbd label legend li link main map mark menu meta meter nav noframes noscript \
object ol optgroup option output p param picture pre progress q rp rt ruby \
s samp script search section select small source span strike strong style sub summary sup svg \
table tbody td template textarea tfoot th thead time title tr track tt u ul var video wbr\
"""
"""source: https://www.w3schools.com/tags/"""


svg_tags = """\
a animate animateMotion animateTransform circle clipPath defs desc ellipse \
feBlend feColorMatrix feComponentTransfer feComposite feConvolveMatrix feDiffuseLighting feDisplacementMap \
feDistantLight feDropShadow feFlood feFuncA feFuncB feFuncG feFuncR feGaussianBlur feImage feMerge feMergeNode \
feMorphology feOffset fePointLight feSpecularLighting feSpotLight feTile feTurbulence filter foreignObject \
g image line linearGradient marker mask metadata mpath path pattern polygon polyline \
radialGradient rect script set stop style svg switch symbol text textPath title tspan use view\
"""
"""source: https://developer.mozilla.org/en-US/docs/Web/SVG/Element"""


all_tags = sorted(set((html_tags + svg_tags).split()))


def generate_elements():
    """generate and write HTML and SVG elements to `elements.py`"""
    with open("ptag/elements.py", "w") as f:
        f.write(f"from .tag import Tag\n")
        f.write(f"from .{Path(__file__).stem} import {typed_partial.__name__}\n\n")
        for tag in all_tags:
            # handle special cases, del is a keyword
            alias = tag + "_" if tag in ["del", "input", "map", "object"] else tag
            f.write(f"{alias} = {typed_partial.__name__}(Tag, '{tag}')\n")


if __name__ == "__main__":
    generate_elements()



from typing import Self
from xml.dom.minidom import Document, parseString
from typing import NamedTuple


class Tag:
    """HTML / SVG element generation class

    Examples:

        - void element, child=None (default)
        >>> Tag('br')
        <br />

        - void Tag w/ attr
        >>> Tag("img", src="http://img.url")
        <img src="http://img.url" />

        - empty string child -> Tag with end tag but no content
        >>> Tag("script", child="", src="url")
        <script src="url"></script>

        - boolean attribute(attribute w/o value)
        >>> Tag("option", "a", "selected")
        <option selected>a</option>

        - mix of boolean attribute and normal attribute
        >>> Tag('a', 'foo', 'm-2', 'rounded', 'text-teal-400', href='bar')
        <a m-2 rounded text-teal-400 href="bar">foo</a>

        - nested Tag
        >>> Tag("div", Tag("i", "x"))
        <div><i>x</i></div>

    Args:
        tag_name (str): tag name
        child (str | Tag | None): inner text or other Tag, default None
        args (list[str], optional): names of value-less attributes
            - eg. `defer`, `selected`
            - useful for UnoCSS attributify mode
        kwargs (dict): tag attributes, in form of `key="val"`

    """

    tag_name: str
    args: list[str] = []
    kwargs: dict[str, str] = {}
    parent: Self | None = None
    children: list[Self | str] | None = None
    context_stack: list[Self] = []  # context stack for context manager

    def __init__(
        self,
        tag_name: str | None,
        child: str | Self | None = None,
        *args,
        **kwargs,
    ):
        """initialize Tag, append to current context stack if exists"""
        self.tag_name = tag_name
        self.args = args
        self.kwargs = kwargs
        self.children = self.prepare_child(child)

        if Tag.context_stack:
            Tag.context_stack[-1].add(self)

    def prepare_child(self, child: str | Self | None) -> list[Self | str] | None:
        """utils, prepare children for the Tag"""
        if child is None:
            return None
        if isinstance(child, Tag):
            child.parent = self
        return [child]

    def __enter__(self):
        """enter context manager, append self to current context stack

        Examples:
            >>> with Tag("div", "hello") as div: # doctest: +SKIP
            ...     Tag("p", "world")
            >>> div                              # doctest: +SKIP
            <div>hello<p>world</p></div>
        """
        Tag.context_stack.append(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """exit context manager, pop self from current context stack"""
        tag = Tag.context_stack.pop()
        # accept children if their parent is itself only
        self.children = [
            child
            for child in self.children
            if isinstance(child, str) or child.parent is self
        ]

    def __repr__(self):
        """string representation of the Tag"""

        # handle python naming restriction
        correct_keys = [
            (
                key.removesuffix("_")
                if key in ("class_", "for_")
                else key.replace("_", "-")
            )
            for key in self.kwargs.keys()
        ]

        kwargs_tuple = zip(correct_keys, self.kwargs.values())

        kwargs = [f'{key}="{val}"' for key, val in kwargs_tuple]
        attr = [self.tag_name] + list(self.args) + kwargs
        attr_str = " ".join(attr)

        content = None if self.children is None else "".join(map(str, self.children))

        # void element with self closing tag
        if content is None:
            return f"<{attr_str} />"

        return f"<{attr_str}>{content}</{self.tag_name}>"

    def add(self, child: str | Self, *args, **kwargs) -> Self:
        """add child and attributes to current Tag

        Examples:
            >>> opt = Tag('option')
            >>> opt.add('hello', 'selected', value='world')
            <option selected value="world">hello</option>
        """
        tail = self.prepare_child(child)
        self.children = tail if self.children is None else self.children + tail
        self.args += args
        self.kwargs.update(kwargs)
        return self

    def prettify(self, indent="    ") -> str:
        """prettify the Tag with indentation"""
        dom = parseString(str(self))
        return dom.childNodes[0].toprettyxml(indent=indent)

    @staticmethod
    def comment(comment: str) -> str:
        """comment element

        Examples:
            >>> Tag.comment("hello")
            '<!--hello-->'
        """
        return Document().createComment(comment).toxml()

    @staticmethod
    def doctype(kind: str = "html") -> str:
        """DOCTYPE element

        Examples:
            >>> Tag.doctype()
            '<!DOCTYPE html>'
        """
        return f"<!DOCTYPE {kind}>"



from .tag import Tag
from .gen_elements import typed_partial

a = typed_partial(Tag, 'a')
abbr = typed_partial(Tag, 'abbr')
acronym = typed_partial(Tag, 'acronym')
address = typed_partial(Tag, 'address')
animate = typed_partial(Tag, 'animate')
animateMotion = typed_partial(Tag, 'animateMotion')
animateTransform = typed_partial(Tag, 'animateTransform')
applet = typed_partial(Tag, 'applet')
area = typed_partial(Tag, 'area')
article = typed_partial(Tag, 'article')
aside = typed_partial(Tag, 'aside')
audio = typed_partial(Tag, 'audio')
b = typed_partial(Tag, 'b')
base = typed_partial(Tag, 'base')
basefont = typed_partial(Tag, 'basefont')
bdi = typed_partial(Tag, 'bdi')
bdo = typed_partial(Tag, 'bdo')
big = typed_partial(Tag, 'big')
blockquote = typed_partial(Tag, 'blockquote')
body = typed_partial(Tag, 'body')
br = typed_partial(Tag, 'br')
button = typed_partial(Tag, 'button')
canvas = typed_partial(Tag, 'canvas')
caption = typed_partial(Tag, 'caption')
center = typed_partial(Tag, 'center')
circle = typed_partial(Tag, 'circle')
cite = typed_partial(Tag, 'cite')
clipPath = typed_partial(Tag, 'clipPath')
code = typed_partial(Tag, 'code')
col = typed_partial(Tag, 'col')
colgroup = typed_partial(Tag, 'colgroup')
data = typed_partial(Tag, 'data')
datalist = typed_partial(Tag, 'datalist')
dd = typed_partial(Tag, 'dd')
defs = typed_partial(Tag, 'defs')
del_ = typed_partial(Tag, 'del')
desc = typed_partial(Tag, 'desc')
details = typed_partial(Tag, 'details')
dfn = typed_partial(Tag, 'dfn')
dialog = typed_partial(Tag, 'dialog')
dir = typed_partial(Tag, 'dir')
div = typed_partial(Tag, 'div')
dl = typed_partial(Tag, 'dl')
dt = typed_partial(Tag, 'dt')
ellipse = typed_partial(Tag, 'ellipse')
em = typed_partial(Tag, 'em')
embed = typed_partial(Tag, 'embed')
feBlend = typed_partial(Tag, 'feBlend')
feColorMatrix = typed_partial(Tag, 'feColorMatrix')
feComponentTransfer = typed_partial(Tag, 'feComponentTransfer')
feComposite = typed_partial(Tag, 'feComposite')
feConvolveMatrix = typed_partial(Tag, 'feConvolveMatrix')
feDiffuseLighting = typed_partial(Tag, 'feDiffuseLighting')
feDisplacementMap = typed_partial(Tag, 'feDisplacementMap')
feDistantLight = typed_partial(Tag, 'feDistantLight')
feDropShadow = typed_partial(Tag, 'feDropShadow')
feFlood = typed_partial(Tag, 'feFlood')
feFuncA = typed_partial(Tag, 'feFuncA')
feFuncB = typed_partial(Tag, 'feFuncB')
feFuncG = typed_partial(Tag, 'feFuncG')
feFuncR = typed_partial(Tag, 'feFuncR')
feGaussianBlur = typed_partial(Tag, 'feGaussianBlur')
feImage = typed_partial(Tag, 'feImage')
feMerge = typed_partial(Tag, 'feMerge')
feMergeNode = typed_partial(Tag, 'feMergeNode')
feMorphology = typed_partial(Tag, 'feMorphology')
feOffset = typed_partial(Tag, 'feOffset')
fePointLight = typed_partial(Tag, 'fePointLight')
feSpecularLighting = typed_partial(Tag, 'feSpecularLighting')
feSpotLight = typed_partial(Tag, 'feSpotLight')
feTile = typed_partial(Tag, 'feTile')
feTurbulence = typed_partial(Tag, 'feTurbulence')
fieldset = typed_partial(Tag, 'fieldset')
figcaption = typed_partial(Tag, 'figcaption')
figure = typed_partial(Tag, 'figure')
filter = typed_partial(Tag, 'filter')
font = typed_partial(Tag, 'font')
footer = typed_partial(Tag, 'footer')
foreignObject = typed_partial(Tag, 'foreignObject')
form = typed_partial(Tag, 'form')
frame = typed_partial(Tag, 'frame')
frameset = typed_partial(Tag, 'frameset')
g = typed_partial(Tag, 'g')
h1 = typed_partial(Tag, 'h1')
h2 = typed_partial(Tag, 'h2')
h3 = typed_partial(Tag, 'h3')
h4 = typed_partial(Tag, 'h4')
h5 = typed_partial(Tag, 'h5')
h6 = typed_partial(Tag, 'h6')
head = typed_partial(Tag, 'head')
header = typed_partial(Tag, 'header')
hgroup = typed_partial(Tag, 'hgroup')
hr = typed_partial(Tag, 'hr')
html = typed_partial(Tag, 'html')
i = typed_partial(Tag, 'i')
iframe = typed_partial(Tag, 'iframe')
image = typed_partial(Tag, 'image')
img = typed_partial(Tag, 'img')
input_ = typed_partial(Tag, 'input')
ins = typed_partial(Tag, 'ins')
kbd = typed_partial(Tag, 'kbd')
label = typed_partial(Tag, 'label')
legend = typed_partial(Tag, 'legend')
li = typed_partial(Tag, 'li')
line = typed_partial(Tag, 'line')
linearGradient = typed_partial(Tag, 'linearGradient')
link = typed_partial(Tag, 'link')
main = typed_partial(Tag, 'main')
map_ = typed_partial(Tag, 'map')
mark = typed_partial(Tag, 'mark')
marker = typed_partial(Tag, 'marker')
mask = typed_partial(Tag, 'mask')
menu = typed_partial(Tag, 'menu')
meta = typed_partial(Tag, 'meta')
metadata = typed_partial(Tag, 'metadata')
meter = typed_partial(Tag, 'meter')
mpath = typed_partial(Tag, 'mpath')
nav = typed_partial(Tag, 'nav')
noframes = typed_partial(Tag, 'noframes')
noscript = typed_partial(Tag, 'noscript')
object_ = typed_partial(Tag, 'object')
ol = typed_partial(Tag, 'ol')
optgroup = typed_partial(Tag, 'optgroup')
option = typed_partial(Tag, 'option')
output = typed_partial(Tag, 'output')
p = typed_partial(Tag, 'p')
param = typed_partial(Tag, 'param')
path = typed_partial(Tag, 'path')
pattern = typed_partial(Tag, 'pattern')
picture = typed_partial(Tag, 'picture')
polygon = typed_partial(Tag, 'polygon')
polyline = typed_partial(Tag, 'polyline')
pre = typed_partial(Tag, 'pre')
progress = typed_partial(Tag, 'progress')
q = typed_partial(Tag, 'q')
radialGradient = typed_partial(Tag, 'radialGradient')
rect = typed_partial(Tag, 'rect')
rp = typed_partial(Tag, 'rp')
rt = typed_partial(Tag, 'rt')
ruby = typed_partial(Tag, 'ruby')
s = typed_partial(Tag, 's')
samp = typed_partial(Tag, 'samp')
script = typed_partial(Tag, 'script')
search = typed_partial(Tag, 'search')
section = typed_partial(Tag, 'section')
select = typed_partial(Tag, 'select')
set = typed_partial(Tag, 'set')
small = typed_partial(Tag, 'small')
source = typed_partial(Tag, 'source')
span = typed_partial(Tag, 'span')
stop = typed_partial(Tag, 'stop')
strike = typed_partial(Tag, 'strike')
strong = typed_partial(Tag, 'strong')
style = typed_partial(Tag, 'style')
sub = typed_partial(Tag, 'sub')
summary = typed_partial(Tag, 'summary')
sup = typed_partial(Tag, 'sup')
svg = typed_partial(Tag, 'svg')
switch = typed_partial(Tag, 'switch')
symbol = typed_partial(Tag, 'symbol')
table = typed_partial(Tag, 'table')
tbody = typed_partial(Tag, 'tbody')
td = typed_partial(Tag, 'td')
template = typed_partial(Tag, 'template')
text = typed_partial(Tag, 'text')
textPath = typed_partial(Tag, 'textPath')
textarea = typed_partial(Tag, 'textarea')
tfoot = typed_partial(Tag, 'tfoot')
th = typed_partial(Tag, 'th')
thead = typed_partial(Tag, 'thead')
time = typed_partial(Tag, 'time')
title = typed_partial(Tag, 'title')
tr = typed_partial(Tag, 'tr')
track = typed_partial(Tag, 'track')
tspan = typed_partial(Tag, 'tspan')
tt = typed_partial(Tag, 'tt')
u = typed_partial(Tag, 'u')
ul = typed_partial(Tag, 'ul')
use = typed_partial(Tag, 'use')
var = typed_partial(Tag, 'var')
video = typed_partial(Tag, 'video')
view = typed_partial(Tag, 'view')
wbra = typed_partial(Tag, 'wbra')
